/* 
        NakView.cc

        MASL/M-MASL Copyright (c) KUNI/TDU$B1~>p8&(B
        Windows version Copyright (c) 1995, 1996 Tomoatsu Shimada  (shimarin)
        X Window System version Copyright (c)1997 Yamahata Isaku (ISA-9)
*/
#include <string.h>

#include "NakView.h"
#include "Scenario.h"

NakView::NakView (const char* const nakFileName,
		  Display* nakDisplay,XMMS_OPTION nakXmmsOption):
  fileName (nakFileName),xmmsOption (nakXmmsOption),display (nakDisplay),
  mmsWindow (NULL),numberOfNak (0),numberOfMmsPalette (0),
  nNak (0),nPalette (0)
{
  /////////////////////////////////////////////////////////$BJQ?t=i4|2=(B
  xmmsOption.isShowVirtualVram=True;
  memset (virtualVram,0,VIRTUAL_VRAM_LENGTH);
  for (int i=0;i<MAXNAK;i++)
    pNak[i]=NULL;
  for (int i=0;i<MAXPAL;i++)
    pMmsPalette[i]=NULL;
  
  /////////////////////////////////////////////////////////X$B4XO"(B
  screen=DefaultScreen (display);
#if 0
  XVisualInfo visualInfo;
  if (!XMatchVisualInfo (display,screen,8,PseudoColor,&visualInfo))
  //if (!XMatchVisualInfo (display,screen,8,DirectColor,&visualInfo))
  //if (!XMatchVisualInfo (display,screen,8,GrayScale,&visualInfo))
    {
      cerr << "this X server does NOT support PseudoColor of 8bpp" << endl;
      exit (-1);
    }
  visual=visualInfo.visual;
  depth=visualInfo.depth;
#endif
  visual=DefaultVisual (display,screen);
  depth=DefaultDepth (display,screen);
  window=XCreateSimpleWindow (display,RootWindow (display,screen),0,0,
                              WINDOW_SIZE_X,WINDOW_SIZE_Y,0,
                              BlackPixel (display,screen),
                              WhitePixel (display,screen));
  ///////////////title$B$r@_Dj!#(B
  char xmms[]="NakView:";
  char* windowTitle=new char[strlen (xmms)+strlen (fileName)+1];
  strcpy (windowTitle,xmms);
  strcat (windowTitle,fileName);
  XStoreName (display,window,windowTitle);
  delete[] windowTitle;

  ////////////////////////////////////////////////////////////$B$=$NB>$N@_Dj(B
  allocColors ();
  XSelectInput (display,window,KeyPressMask|ExposureMask|ButtonPressMask|
		ButtonReleaseMask|Button1MotionMask);
  XMapWindow (display,window);
  XFlush (display);
  XEvent event;
  do
    {//$B2hLL$,(Bmap$B$5$l$k$^$GBT$D(B
      XNextEvent (display,&event);
    }while (event.xexpose.count!=0);

  ///////////////////////////////////////////////////////////file$B$N(Bload
  if (!loadNak (fileName))
    {
      cerr << "can not load " << fileName << endl;
      cerr << "Is " << fileName << " broken?" << endl;
      delete mmsWindow;
      for (int i=0;i<MAXNAK;i++)delete pNak[i];
      for (int i=0;i<MAXPAL;i++)delete pMmsPalette[i];
      XFreeGC (display,gc);
      if (colormap!=XDefaultColormap (display,screen))
	XFreeColormap (display,colormap);
      XDestroyWindow (display,window);
      return;
    }
}


NakView::~NakView ()
{
  delete mmsWindow;
  for (int i=0;i<MAXNAK;i++)delete pNak[i];
  for (int i=0;i<MAXPAL;i++)delete pMmsPalette[i];
  XFreeGC (display,gc);
  if (colormap!=XDefaultColormap (display,screen))
    XFreeColormap (display,colormap);
  XDestroyWindow (display,window);
}

//////////////////////////////////////////////////////////Nak$B$N(Bload
Bool NakView::loadNak (const char* const fileName)
{
  cout << "now loading " << fileName << "." << endl;
  ////////////////////////////////////////////////////////file$B$N(Bopen
  MaslInputStream maslIs (fileName,ios::in|ios::nocreate);
  if (!maslIs)
    {
      cerr << fileName << "::File couldn't open.";
      perror ("");
      return False;
    }
  const int fileSize=maslIs.getFileSize ();

  ///////////////////////////////////////////////////magic number$B$N(Bcheck
  char nak[9];
  maslIs.read (nak,8);
  nak[8]='\0';
  if (strncmp (nak,"NAK",3)!=0)
    {
      cerr << "Incorrect data format detected." << endl;
      return False;
    }
       
  //$B$3$NNN0h$O2?$r$7$F$$$k$+ITL@(B(comment$BNN0h$+!)(B)
  char auther[33];
  auther[32]='\0';
  maslIs.read (auther,32);
  cout << "auther=";
  Scenario::writeSjis (cout,auther,strlen (auther),xmmsOption.isSjis);
  cout << endl;
  int tmpsize=maslIs.get16 ();
  char comment[tmpsize];
  maslIs.read (comment,tmpsize);
  cout << "comment=";
  Scenario::writeSjis (cout,comment,tmpsize,xmmsOption.isSjis);
  cout << endl;
  cout << "id:0=palette,1=normal save,2=clear save,3=diff save" << endl;
 
  ///////////////////////////////////////////////////////data$BFI$_9~$_(B
  while (maslIs.tellg ()<fileSize)
    {
      unsigned char nakId=maslIs.get8 ();
      cout << "id=" << (int) nakId << ":";
      
      switch (nakId)
        {
        case 0://palette data$BFI$_9~$_(B
          pMmsPalette[numberOfMmsPalette++]=new MmsPalette(maslIs);
	  cout << numberOfMmsPalette << "-th palette data" << endl;
          break;
          
        case 1://$BD>2<$X(B
        case 2://$BD>2<$X(B
        case 3://$B2hA|(Bdata$BFI$_9~$_(B
	  pNak[numberOfNak++]=new Nak (maslIs,nakId);
	  cout  << numberOfNak << "-th graphic data" << endl;
          break;
          
        default://$BDL>o$O$3$3$X$OMh$J$$!#(B
          cerr << "Incorrect NAk ID detected" << endl;
          return False;
        }
    }
  cout << numberOfMmsPalette << " palettes,"
       << numberOfNak << " graphics" << endl;
  return True;
}

/////////////////////////////////////$B?'$N3NJ]Ey$O(Bvisual class$B$K0MB8$9$k!#(B
void NakView::allocColors ()
{
  colormap=XDefaultColormap (display,screen);
  if (xmmsOption.isPrivateColormapCreated ||
      XAllocColorCells (display,colormap,False,
                        planeMask,DEPTH_OF_NAK,&basePixel,1)!=True)
    {//option$B;XDj$5$l$F$$$k$+(Bcommon colormap$B$+$i?'$r3NJ]$G$-$J$1$l$P(B
      //private colormap$B$r;HMQ$9$k!#(B
      if (xmmsOption.isPrivateColormapCreated==False)
        {
          cerr << "can not allocate colors! not enough color.\n";
          cerr << "creating private color map." << endl;
        }
      xmmsOption.isPrivateColormapCreated=True;
      colormap=XCreateColormap (display,RootWindow (display,screen),
                                visual,AllocAll);  
      XInstallColormap (display,colormap);
      XSetWindowColormap (display,window,colormap);
      basePixel=0L;
      for (int i=0;i<DEPTH_OF_NAK;i++)
        planeMask[i]=(1L<<i);
    }

   for (int i=0;i<NUM_OF_COLORS;i++)
    {
      palette2Pixel[i]=basePixel;
      for (int j=0;j<DEPTH_OF_NAK;j++)
        palette2Pixel[i]|= ((i&(1 << j))?planeMask[j]:0);
    }

  gc=XCreateGC (display,window,0,0);
  XSetWindowBackground (display,window,BlackPixel (display,screen));

  //MmsWindow$B$H(BMmsPaltte$B$KI,MW$J>pJs$rEO$9!#(B
  mmsWindow=new MmsWindow (display,screen,visual,gc,depth,window,
                           palette2Pixel,
                           &xmmsOption);
  MmsPalette::prepareMmsPalette (display,colormap,palette2Pixel);
}


////////////////////////////////////////////////public method
int NakView::putImage (const int n)
{
  nNak=abs (n)%numberOfNak;
  pNak[nNak]->extractNak2Vram (virtualVram,0,0);
  mmsWindow->putImage (virtualVram);
  XFlush (display);
  return nNak;
}

int NakView::setPalette (int n)
{
  nPalette=abs (n)%numberOfMmsPalette;
  pMmsPalette[nPalette]->setPalette (0);
  mmsWindow->putImageOfVirtualWindow (virtualVram);
  XFlush (display);
  return nPalette;
}

void NakView::clearWindow ()
{
  memset (virtualVram,0,VIRTUAL_VRAM_LENGTH);
  mmsWindow->putImageOfVirtualWindow (virtualVram);
  XFlush (display);
}
